/*
 * FILE NAME ralink_gpio_cp.c
 *
 * BRIEF MODULE DESCRIPTION
 *  Driver for Ralink RT3XXX GPIO, with different functionality than Ralink's version.
 *
 *  Author: CradlePoint Technology, Inc.  <source@cradlepoint.com>
 *          Steve Johnson <stevej@cradlepoint.com>
 *
 *  Derived from Steve Longerbeam's au1000_gpio.c
 *  Copyright 2001 MontaVista Software Inc.
 *
 *  Copyright 2012 CradlePoint Technology, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2
 *  as published by the Free Software Foundation.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE	LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/kobject.h>
#include <linux/miscdevice.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/jiffies.h>
#include <linux/timer.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/bitops.h>
#include <asm/mach-ralink/cpboard.h>
#include "ralink_gpio_cp.h"

/*
  Keep track of some flags around each pin.  Is it blinking?
*/
//static struct timer_list blink_timer;
static char blink_array[GPIO_PINS];

static int ralink_gpio_blink_start(u32 pin);
static int ralink_gpio_blink_stop(u32 pin);
static int ralink_gpio_serial_enable(void);
static int ralink_gpio_serial_disable(void);
static void ralink_gpio_blink_process(unsigned long unused);
DEFINE_TIMER(blink_timer, ralink_gpio_blink_process, 0, 0);

/*
  Interrupt GPIO pin debounce handling.
*/
#define DEBOUNCE_TIME 200   // mSec
static void debounce_process(unsigned long data);
DEFINE_TIMER(debounce_timer, debounce_process, 0, 0);
static volatile long unsigned int disabled_pin[3];   // Needs to hold up to 96 GPIOs

#define GPIO_ENTRIES (sizeof(ralink_gpio_reg)/sizeof(ralink_gpio_reg[0]))

#define GPIO_INTERRUPT 14

//#define DEBUG 1
#undef DEBUG

static void __iomem *base_addr;

/*
  These allow for the kernel event to be sent from the
  interrupt handling code.
*/
static struct kset *gpio_kset;
static struct kobject *gpio_kobj;
static struct work_struct gpio_event;

/*
 * Return the pin state. Pins configured as outputs will return
 * the output state, and pins configured as inputs (tri-stated)
 * will return input pin state.
 */
int ralink_gpio_read(u32 pin)
{
	int i = 0;
	int offset = 0;
	
	if (pin > UPGRADE) {
		printk(KERN_WARNING "Reading from unmapped GPIO pin\n");
		return -EFAULT;
	}

	// Special check, this is a command not a pin
	if (BOOT_COUNT <= pin) {
		//prink(KERN_INFO "Pin found %d\n", pin);
		switch(pin)
			{
			case BOOT_COUNT:
				if (1 == rt3xxx_check_memo_reg(NO_SECURITY)) {
					//prink(KERN_INFO "found no security\n");
					return 1;
				} else {
					//prink(KERN_INFO "found security\n");
					return 0;
				}
				break;
			}
		return 0;
	}

	while (pin >= ralink_gpio_reg[i].max) {
		offset = ralink_gpio_reg[i].max;
		i++;
	}

#ifdef DEBUG1
	printk(KERN_INFO "offset %d, dir 0x%x, data 0x%x\n", offset, ralink_gpio_reg[i].dir, ralink_gpio_reg[i].data);
	printk(KERN_INFO "Dir reg 0x%x\n", READ_PIO(ralink_gpio_reg[i].dir));
	printk(KERN_INFO "Data reg 0x%x\n", READ_PIO(ralink_gpio_reg[i].data));
#endif 

	// return pin value
	if ((1 << (pin - offset)) & READ_PIO(ralink_gpio_reg[i].data)) {
		return 1;
	} else {
		return 0;
	}
}


/*
 * Set/clear GPIO pins. Only available GPIOs will be affected.
 */
int ralink_gpio_write(u32 pin, u32 value)
{

#define MAGIC	0x27051956
#define UPGRADE_MAGIC	0x27051957
	int i = 0;
	int offset = 0;
	
#ifdef DEBUG
	printk(KERN_INFO "%s, pin %d, val = %d\n", __FUNCTION__, pin, value);
#endif

	if (pin > UPGRADE) {
		printk(KERN_WARNING "Writing to unmapped GPIO pin\n");
		return -EFAULT;
	}

	// Special check, is this a command not a pin?
	if (BOOT_COUNT <= pin) {
#ifdef DEBUG
		printk(KERN_INFO "%s Special pin found %d\n", __FUNCTION__, pin);
#endif
		switch(pin)
			{
			case BOOT_COUNT:
				// Reset boot count
				if (MAGIC == READ_PIO(RALINK_SYSCTL_MEMO0)) {
					WRITE_PIO(0, RALINK_SYSCTL_MEMO0);
				}
				break;
			case UPGRADE:
				// Set upgrade flag
				WRITE_PIO(UPGRADE_MAGIC, RALINK_SYSCTL_MEMO0);
				WRITE_PIO(UPGRADE_MAGIC, RALINK_SYSCTL_MEMO0);
				break;
			}
		return 0;
	}

	while (pin >= ralink_gpio_reg[i].max) {
		offset = ralink_gpio_reg[i].max;
		i++;
	}

	// Is the pin mapped as an output?
	if (((1 << (pin - offset)) & READ_PIO(ralink_gpio_reg[i].dir)) == 0) {
		printk(KERN_DEBUG "Writing to an input GPIO pin %d, offset = %d, 0x%x\n", pin, offset, READ_PIO(ralink_gpio_reg[i].dir));
		return -EFAULT;
	}

#ifdef DEBUG1
	if (value) {
		printk(KERN_INFO "Old value is 0x%x\n", READ_PIO(ralink_gpio_reg[i].data));
		printk(KERN_INFO "Writing 0x%x to 0x%p\n", cpu_to_le32(1 << (pin - offset)), base_addr + ralink_gpio_reg[i].set);
		WRITE_PIO((1 << (pin - offset)), ralink_gpio_reg[i].set);
		printk(KERN_INFO "New value is 0x%x\n", READ_PIO(ralink_gpio_reg[i].data));
	} else {
		printk(KERN_INFO "Old value is 0x%x\n", READ_PIO(ralink_gpio_reg[i].data));
		printk(KERN_INFO "Writing 0x%x to 0x%p\n", cpu_to_le32(1 << (pin - offset)), base_addr + ralink_gpio_reg[i].clear);
		WRITE_PIO((1 << (pin - offset)), ralink_gpio_reg[i].clear);
		printk(KERN_INFO "New value is 0x%x\n", READ_PIO(ralink_gpio_reg[i].data));
	}
#else

	if (value) {
		WRITE_PIO((1 << (pin - offset)), ralink_gpio_reg[i].set);
	} else {
		WRITE_PIO((1 << (pin - offset)), ralink_gpio_reg[i].clear);
	}

#endif
	return 0;
}

int ralink_gpio_dir_read(u32 pin)
{
	
	int offset = 0;
	int i = 0;
	int regval;

	if (pin > UPGRADE) {
		printk(KERN_WARNING "Writing to unmapped GPIO pin\n");
		return -EFAULT;
	}

	while (pin >= ralink_gpio_reg[i].max) {
		offset = ralink_gpio_reg[i].max;
		i++;
	}
	regval = READ_PIO(ralink_gpio_reg[i].dir) | (1 << (pin - offset));
	regval = (regval >> pin) & 0x1;

	return regval; 
}


int ralink_gpio_dir_write(u32 pin, u32 value)
{
	int offset = 0;
	int i = 0;

	if (pin > UPGRADE) {
		printk(KERN_WARNING "Writing to unmapped GPIO pin\n");
		return -EFAULT;
	}

	while (pin >= ralink_gpio_reg[i].max) {
		offset = ralink_gpio_reg[i].max;
		i++;
	}

	int regvaldir = READ_PIO(ralink_gpio_reg[i].dir);
	if (value == 0) {
		WRITE_PIO(regvaldir & ~(0x1 << (pin - offset)), ralink_gpio_reg[i].dir);
	} else {
		WRITE_PIO(regvaldir | (0x1 << (pin - offset)), ralink_gpio_reg[i].dir);
	}
	return 0;
}


EXPORT_SYMBOL(ralink_gpio_read);
EXPORT_SYMBOL(ralink_gpio_write);
EXPORT_SYMBOL(ralink_gpio_dir_read);
EXPORT_SYMBOL(ralink_gpio_dir_write);


static int ralink_gpio_open(struct inode *inode, struct file *file)
{
	try_module_get(THIS_MODULE);
	return 0;
}


static int ralink_gpio_release(struct inode *inode, struct file *file)
{
	module_put(THIS_MODULE);
	return 0;
}

static int ralink_gpio_blink_start(u32 pin)
{
	int i = 0;
	int offset = 0;
	
#ifdef DEBUG
	printk(KERN_INFO "%s, pin %d\n", __FUNCTION__, pin);
#endif	

	if (pin > UPGRADE) {
		printk(KERN_INFO "Blinking unmapped GPIO pin\n");
		return -EFAULT;
	}

	while (pin >= ralink_gpio_reg[i].max) {
		offset = ralink_gpio_reg[i].max;
		i++;
	}

	// Is the pin mapped as an output?
	if (((1 << (pin - offset)) & le32_to_cpu(READ_PIO(ralink_gpio_reg[i].dir))) == 0) {
		printk(KERN_INFO "Writing to an input GPIO pin\n");
		return -EFAULT;
	}

	blink_array[pin] = 1;

	// Check blinking thread.  If one doesn't exist,
	//  start one.
	if (timer_pending(&blink_timer) == 0) {
		// Create a timer
#ifdef DEBUG
		printk(KERN_INFO "Create new blink timer\n");
#endif
		mod_timer(&blink_timer, jiffies + (HZ / 2));
	}

	return 0;
}

static int ralink_gpio_blink_stop(u32 pin)
{
	int i = 0;
	int offset = 0;
	
#ifdef DEBUG
	printk(KERN_INFO "%s, pin %d\n", __FUNCTION__, pin);
#endif	

	if (pin > UPGRADE) {
		printk(KERN_INFO "Blinking unmapped GPIO pin\n");
		return -EFAULT;
	}

	while (pin >= ralink_gpio_reg[i].max) {
		offset = ralink_gpio_reg[i].max;
		i++;
	}

	// Is the pin mapped as an output?
	if (((1 << (pin - offset)) & le32_to_cpu(READ_PIO(ralink_gpio_reg[i].dir))) == 0) {
		printk(KERN_INFO "Writing to an input GPIO pin\n");
		return -EFAULT;
	}

	if (blink_array[pin] == 0) {
		// This can happen with normal processing
		// printk(KERN_INFO "Disabling a pin that isn't blinking\n");
		return 0;		
	}

	blink_array[pin] = 0;

	// Make sure the pin is set to off, don't want to leave it stuck on.
	ralink_gpio_write(pin, 1);

	return 0;
}

/*
 *  If any LED is set to blink, make sure they all blink in
 * sequence.  It looks funny if they don't blink together.
 * If there are any blinking LEDs, reschedule every half
 * second.  
 */
static void ralink_gpio_blink_process(unsigned long unused)
{
	int pin;
	int reschedule = 0;
	static int blink_on = 0;

#ifdef DEBUG1
	printk(KERN_INFO "%s\n", __FUNCTION__);
#endif

	for (pin = 0; pin < GPIO_PINS; pin++) {
		if (blink_array[pin]) {
			ralink_gpio_write(pin, blink_on);
			reschedule = 1;
		}
	}

	// Toggle blink
	if (blink_on) {
		blink_on = 0;
	} else {
		blink_on = 1;
	}

	if (reschedule) {
		mod_timer(&blink_timer, jiffies + (HZ / 2));
	} else {
		// reset to default
		blink_on = 0;
	}

	return;
}
static int ralink_gpio_serial_enable()
{
	int reg = READ_PIO(RALINK_REG_GPIOMODE);
	reg |= (7 << 2); // set UARTF_SHARE_MODE to GPIO mode
	WRITE_PIO(reg, RALINK_REG_GPIOMODE);

	return 0;
}
static int ralink_gpio_serial_disable()
{
	int reg = READ_PIO(RALINK_REG_GPIOMODE);
	reg &= ~(7 << 3); // set UARTF_SHARE_MODE to UART mode
	WRITE_PIO(reg, RALINK_REG_GPIOMODE);

	return 0;
}

static long ralink_gpio_ioctl(struct file *file,
							  unsigned int cmd, unsigned long arg)
{
#ifdef DEBUG
	printk(KERN_INFO "%s, cmd = 0x%x, val = 0x%lx\n", __FUNCTION__, cmd, arg);
#endif

	switch(cmd) {
	case RALINK_GPIO_READ_BIT:
		
		return ralink_gpio_read(arg);

	case RALINK_GPIO_SET_BIT:
		{
			return ralink_gpio_write(arg /* pin */, 1 /* value */);
		}

	case RALINK_GPIO_CLEAR_BIT:
		{
			return ralink_gpio_write(arg /* pin */, 0 /* value */);
		}

	case RALINK_GPIO_READ_DIR_BIT:
		
		return ralink_gpio_dir_read(arg);

	case RALINK_GPIO_SET_DIR_BIT:
		{
			return ralink_gpio_dir_write(arg /* pin */, 1 /* value */);
		}

	case RALINK_GPIO_CLEAR_DIR_BIT:
		{
			return ralink_gpio_dir_write(arg /* pin */, 0 /* value */);
		}

	case RALINK_GPIO_BLINK:

		return ralink_gpio_blink_start(arg);
		
	case RALINK_GPIO_NO_BLINK:

		return ralink_gpio_blink_stop(arg);

	case RALINK_GPIO_SERIAL_ENABLE:
		{
			return ralink_gpio_serial_enable();
		}
	
	case RALINK_GPIO_SERIAL_DISABLE:
		{
			return ralink_gpio_serial_disable();
		}
	default:
		return -ENOIOCTLCMD;

	}

	return 0;
}

static void gpio_send_event(struct work_struct *ws)
{
	unsigned long flags;
	int i;
	char event_string[32];
	char *event[] = { &event_string[0], NULL }; 

#ifdef DEBUG
	printk(KERN_INFO "%s\n", __FUNCTION__);
#endif

	for (i=0; i < DEBOUNCED_PINS; i++) {
		if (ralink_gpio_event[i].gpio_event_occurred) {
#if 0
			printk(KERN_INFO "event_occurred count: %d\n", ralink_gpio_event[i].gpio_event_occurred);
#endif
			local_irq_save(flags);
			ralink_gpio_event[i].gpio_event_occurred--;
			local_irq_restore(flags);

			sprintf(&event_string[0], "GPIO %d, event %d\n", 
					ralink_gpio_event[i].gpio_interrupt, 
					ralink_gpio_event[i].gpio_event);

			if (gpio_kobj) {
				kobject_uevent_env(gpio_kobj, KOBJ_CHANGE, event);
			}
#ifdef DEBUG
			printk(KERN_INFO "%s", event_string);
#endif
		}
	}

}

static void enable_gpio_pin(int pin, int bank)
{
	int temp;
	unsigned long flags;

#ifdef DEBUG
	printk(KERN_INFO "%s pin %d, bank %d\n", __FUNCTION__, pin, bank);
#endif

	local_irq_save(flags);
	if (ralink_gpio_reg[bank].rena_value & (1 << pin)) {
		temp = READ_PIO(ralink_gpio_reg[bank].int_rena) | (1 << pin);
		WRITE_PIO(temp, ralink_gpio_reg[bank].int_rena); 
	}
	if (ralink_gpio_reg[bank].fena_value & (1 << pin)) {
		temp = READ_PIO(ralink_gpio_reg[bank].int_fena) | (1 << pin);
		WRITE_PIO(temp, ralink_gpio_reg[bank].int_fena); 
	}
	local_irq_restore(flags);
}

static void debounce_process(unsigned long data)
{
	int bank;
	int offset = 0;
	int pin = 0;
	int i;
	unsigned long flags;

#ifdef DEBUG
	printk(KERN_INFO "%s\n", __FUNCTION__);
#endif

	local_irq_save(flags);
	for (i=0; i < 3; i++) {
		while (disabled_pin[i]) {
			// This gives the pin location within the word
			pin = fls(disabled_pin[i]) - 1;
			clear_bit(pin, &disabled_pin[i]);

			// This is the real pin number
			pin += (BITS_PER_LONG * i);

			bank = 0;
			offset = 0;
			while (pin >= ralink_gpio_reg[bank].max) {
				offset = ralink_gpio_reg[bank].max;
				bank++;
			}
			pin -= offset;
			enable_gpio_pin(pin, bank);
		}
	}
	local_irq_restore(flags);

#ifdef DEBUG
	printk(KERN_INFO "Interrupt pin %d reenabled\n", pin + offset);
#endif

}


static void disable_gpio_pin(int pin, int bank, int offset)
{
	volatile long unsigned int temp;

#ifdef DEBUG
	printk(KERN_INFO "%s pin %d, bank %d\n", __FUNCTION__, pin, bank);
#endif

	temp = READ_PIO(ralink_gpio_reg[bank].int_rena);
	if (test_and_clear_bit(pin, &temp)) {
		WRITE_PIO(temp, ralink_gpio_reg[bank].int_rena); 
	}

	temp = READ_PIO(ralink_gpio_reg[bank].int_fena);
	if (test_and_clear_bit(pin, &temp)) {
		WRITE_PIO(temp, ralink_gpio_reg[bank].int_fena); 
	}
}


static void ralink_gpio_debounce_setup(void)
{
	int i, j, pin;
	int offset = 0;

#ifdef DEBUG
	printk(KERN_INFO "%s\n", __FUNCTION__);
#endif

	for (i=0; i < GPIO_ENTRIES; i++) {
		// Do we have an interrupt in this bank?
		if (ralink_gpio_reg[i].intp_value) {
		    // Go through the bank, checking for pins, resetting
			//  them to 0, and marking them as disabled
			while(ralink_gpio_reg[i].intp_value) {
				pin = fls(ralink_gpio_reg[i].intp_value) - 1;
				clear_bit(pin, (volatile long unsigned int *)&ralink_gpio_reg[i].intp_value);
				disable_gpio_pin(pin, i, offset);
				set_bit((pin + offset), &disabled_pin[0]); 

				// Handle overcurrent and GPIO events
				for (j=0; j < DEBOUNCED_PINS; j++) {
					if (ralink_gpio_event[j].gpio_interrupt == (pin + offset)) {
						// Check for overcurrent interrupts, if one is
						//  received, then we need to power off the USB port
						if (ralink_gpio_event[j].gpio_disable) {
							ralink_gpio_write(ralink_gpio_event[j].gpio_disable, 0);
						}

						// Mark an event
						ralink_gpio_event[j].gpio_event_occurred++;
					}
				}
			}
		}
		offset = ralink_gpio_reg[i].max;
	}

	// Start a debounce timer
	//  The Reset pin rings like a bell.  Keep it disabled longer
	if (test_bit(SOFT_RST_IN_BUTTON, &disabled_pin[0])) {
		printk(KERN_INFO "Reset button pushed\n");
		mod_timer(&debounce_timer, jiffies + (3 * msecs_to_jiffies(DEBOUNCE_TIME)));
	} else {
		mod_timer(&debounce_timer, jiffies + msecs_to_jiffies(DEBOUNCE_TIME));
	}

	// Send a kernel uevent
	schedule_work(&gpio_event);
	
}


/*
 * 1. save the PIOINT and PIOEDGE value
 * 2. clear PIOINT by writing 1
 * (called by interrupt handler)
 */
void ralink_gpio_save_clear_intp(void)
{

	int i;

	for (i=0; i < GPIO_ENTRIES; i++) {
		// Store interrupt value
		ralink_gpio_reg[i].intp_value = READ_PIO(ralink_gpio_reg[i].interrupt);

		// Clear interrupts

		WRITE_PIO(ralink_gpio_reg[i].intp_value, ralink_gpio_reg[i].interrupt);
		WRITE_PIO(ralink_gpio_reg[i].intp_value, ralink_gpio_reg[i].int_edge);

	}

}

irqreturn_t ralink_gpio_irq_handler(int irq, void *irqaction)
{
#ifdef DEBUG
	printk(KERN_INFO "Received GPIO interrupt\n");
#endif

	ralink_gpio_save_clear_intp();
	ralink_gpio_debounce_setup();

	return IRQ_HANDLED;
}

static struct file_operations ralink_gpio_fops =
{
	.owner		= THIS_MODULE,
	.unlocked_ioctl		= ralink_gpio_ioctl,
	.open		= ralink_gpio_open,
	.release	= ralink_gpio_release,
};


static struct miscdevice ralink_gpio_miscdev =
{
	MISC_DYNAMIC_MINOR,
	"ralink_gpio",
	&ralink_gpio_fops
};


void __exit ralink_gpio_exit(void)
{
	free_irq(GPIO_INTERRUPT, ralink_gpio_irq_handler);
	misc_deregister(&ralink_gpio_miscdev);
	iounmap(base_addr);
	kobject_put(gpio_kobj);
}


int __init ralink_gpio_init(void)
{
	int i, err;

	printk(KERN_INFO "Ralink GPIO driver init.\n");

	err = misc_register(&ralink_gpio_miscdev);
	if (err) {
		printk(KERN_INFO "GPIO driver couldn't register misc device\n");
		return err;
	}

	base_addr = ioremap(RALINK_SYSCTL_ADDR, 0x1000);
	if (!base_addr) {
		panic("Could not ioremap gpio I/O port range");
	}

	/* Save off some register values for later */
	for (i = 0; i < GPIO_ENTRIES; i++) {
		ralink_gpio_reg[i].rena_value = READ_PIO(ralink_gpio_reg[i].int_rena);
		ralink_gpio_reg[i].fena_value = READ_PIO(ralink_gpio_reg[i].int_fena);
	}

	/* Kind of useless, but necessary to send a uevent */
	gpio_kset = kset_create_and_add("gpio_kset", NULL, kernel_kobj);

	/* Create a kobject to use to send events later */
	gpio_kobj = kobject_create();
	if (!gpio_kobj) {
		err = -ENOMEM;
		goto out_err;
	}
	gpio_kobj->kset = gpio_kset;
	if ((err = kobject_add(gpio_kobj, kernel_kobj, "%s", "ralink_gpio_int"))) {
		goto out_err;
	}

	/* Create a work queue to send events.  Can't send an event in an
	 * interrupt context or timer */
	INIT_WORK(&gpio_event, gpio_send_event);

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,35)
	err = request_irq(GPIO_INTERRUPT, ralink_gpio_irq_handler, \
			IRQF_SHARED, "ralink_gpio", ralink_gpio_irq_handler);
#else
	err = request_irq(GPIO_INTERRUPT, ralink_gpio_irq_handler, \
			SA_INTERRUPT, "ralink_gpio", NULL);
#endif
	if (err) {
		printk(KERN_INFO "GPIO driver couldn't attach to interrupt\n");
		goto out_err;
	}

#ifdef DEBUG1
	printk(KERN_INFO "GPIOMODE=0x%x\n", READ_PIO(RALINK_REG_GPIOMODE));

	for (i = 0x600; i < LAST_REG; i += 4) {
		printk(KERN_INFO "GPIO(%x)=0x%x\n", i, READ_PIO(i));
	}
	printk(KERN_INFO "\n");

	printk(KERN_INFO "READ_BIT 0x%x\n", RALINK_GPIO_READ_BIT);
	printk(KERN_INFO "SET_BIT 0x%x\n", RALINK_GPIO_SET_BIT);
	printk(KERN_INFO "CLEAR_BIT 0x%x\n", RALINK_GPIO_CLEAR_BIT);
	printk(KERN_INFO "BLINK 0x%x\n", RALINK_GPIO_BLINK);
	printk(KERN_INFO "NO_BLINK 0x%x\n", RALINK_GPIO_NO_BLINK);

	/*ralink_gpio_write(GPIO_PINS, 1);*/
	printk(KERN_INFO "GPIO write to pin %d needs to be removed once app space is running\n", GPIO_PINS);
#endif

	return 0;

out_err:
	misc_deregister(&ralink_gpio_miscdev);
	iounmap(base_addr);
	kobject_put(gpio_kobj);

	return err;
}


module_init(ralink_gpio_init);
module_exit(ralink_gpio_exit);


/*

// Unit Tests
cppython
import array, fcntl, struct, termios, os
fd = open("/dev/ralink_gpio", "w")
leds = [72, 73, 80, 81]
for i in range(len(leds)):
	buf = leds[i]
	fcntl.ioctl(fd.fileno(), 0x80044702, buf)  // set

>>> for i in range(len(leds)):
...     buf = leds[i]
...     fcntl.ioctl(fd.fileno(), 0x80044703, buf)  // reset

>>> for i in range(96):
...     try:
...             fcntl.ioctl(fd.fileno(), 0x80044702, i)  // set all gpios
...     except:
...             pass

>>> for i in range(96):
...     try:
...             fcntl.ioctl(fd.fileno(), 0x80044703, i)   // reset all gpios (turn on all leds)
...     except:
...             pass
... 

>>> for i in range(96):
...     try:
...             val = fcntl.ioctl(fd.fileno(), 0x40044701, i)  // read all gpios
...             print(val)
...     except:
...             pass

*/
